library std;
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity interface_vga is
  port(
    clock_25Mhz  : in  std_logic;
    reset_n      : in  std_logic;                     -- reset

    ---------------------------------------------------------------------------
    -- Ports of Handshake
    ---------------------------------------------------------------------------
    in_request    : out std_logic;
    in_valid      : in  std_logic;      -- in_data
    in_data       : in  signed(31 downto 0);

    -- Remover depois
    out_valid     : out std_logic;
    ---------------------------------------------------------------------------
    -- Ports of video controller
    ---------------------------------------------------------------------------


    red	         : out std_logic_vector(7 downto 0);  -- red component
    green	 : out std_logic_vector(7 downto 0);  -- green component
    blue	 : out std_logic_vector(7 downto 0);  -- blue component
    line_pixel   : out std_logic_vector(9 downto 0);  -- compute line
    column_pixel : out std_logic_vector(9 downto 0);  -- compute column
    m1, m2       : out std_logic;                     -- select dac mode
    blank_n      : out std_logic;                     -- dac command
    sync_n       : out std_logic;                     -- dac command
    sync_t       : out std_logic;                     -- dac command
    video_clk    : out std_logic;                     -- dac command
    vga_vs       : out std_logic;                     -- vertical sync
    vga_hs       : out std_logic                      -- horizontal sync
    );
end interface_vga;


architecture a_lancelot of interface_vga is
  component tsyncram
    generic (
      DEPTH : positive := 64;             -- How many slots
      DATA_WIDTH : positive := 16;        -- How many bits per slot
      ADDR_WIDTH : positive := 6          -- = ceil(log2(DEPTH))
      );
    port (
      clk : in std_logic;
      wr_e  : in std_logic;
      wr_addr : in unsigned(ADDR_WIDTH-1 downto 0);
      wr_data : in signed(DATA_WIDTH-1 downto 0);
      rd_addr : in unsigned(ADDR_WIDTH-1 downto 0);
      rd_data : out signed(DATA_WIDTH-1 downto 0)
      );
  end component tsyncram;
  
  constant HORIZ_RES           : natural := 640;  -- Horizontal Resolution
  constant HSYNC_BACK_PORCH_W  : natural := 40;
  constant HSYNC_FRONT_PORCH_W : natural := 25;
  constant HSYNC_PULSE_W   : natural := 95;
  constant VERT_RES            : natural := 480;  -- Vertical Resolution
  constant VSYNC_BACK_PORCH_W  : natural := 22;
  constant VSYNC_FRONT_PORCH_W : natural := 10;
  constant VSYNC_PULSE_W       : natural := 2;

--   constant HORIZ_RES           : natural := 160;  -- Horizontal Resolution
--   constant HSYNC_BACK_PORCH_W  : natural := 40;
--   constant HSYNC_FRONT_PORCH_W : natural := 25;
--   constant HSYNC_PULSE_W   : natural := 95;
--   constant VERT_RES            : natural := 120;  -- Vertical Resolution
--   constant VSYNC_BACK_PORCH_W  : natural := 22;
--   constant VSYNC_FRONT_PORCH_W : natural := 10;
--   constant VSYNC_PULSE_W       : natural := 2;

  
-- DAC State Machine
  type dac_mode_state_type is 
    (
      ds0, ds1, ds2, ds3, ds4
      );
  signal dac_mode_current_state, dac_mode_next_state : dac_mode_state_type;
 
  signal count_h, count_v : natural RANGE 0 TO 1023;
  signal video_on_v, video_on_h : std_logic;
  

  signal mem_read_line : integer range 0 to 1023;
  signal mem_read_col  : integer range 0 to 1023;


  -----------------------------------------------------------------------------
  -- Video parameter signals
  -----------------------------------------------------------------------------
  signal video_width, video_height : integer range 0 to 1023;

  -----------------------------------------------------------------------------
  -- Signals of Read process
  -----------------------------------------------------------------------------
  signal counter_cols, counter_lines : integer range 0 to 1023;
  type read_state_t is (stt_read_height,
                        stt_read_width,
                        stt_read_Y,
                        stt_read_Cb,
                        stt_read_Cr
    );
  signal read_state : read_state_t;

  -----------------------------------------------------------------------------
  -- Signal of write process
  -----------------------------------------------------------------------------
  signal pixel_col, pixel_line : integer range 0 to 1023;
  type write_state_t is (stt_write_data, stt_write_wait1,
                         stt_write_wait2, stt_write_wait3,
                         stt_write_none
                         );
  signal write_state : write_state_t;
  
  -----------------------------------------------------------------------------
  -- Buffer memories parameters
  -----------------------------------------------------------------------------
  constant MEM_Y_DEPTH : natural := 1920; --4800;
  constant MEM_Y_DATA_WIDTH : natural := 32;
  constant MEM_Y_ADDR_WIDTH : natural := 11; --13;


  constant MEM_DEPTH : natural := 960;--2400;
  constant MEM_DATA_WIDTH : natural := 32;
  constant MEM_ADDR_WIDTH : natural := 10;--12;

  -----------------------------------------------------------------------------
  -- signals of Y memory
  -----------------------------------------------------------------------------
  signal mem_Y_wr_e     : std_logic;
  signal mem_Y_wr_addr  : unsigned (MEM_Y_ADDR_WIDTH-1 downto 0);
  signal mem_Y_wr_data  : signed   (MEM_Y_DATA_WIDTH-1 downto 0);
  signal mem_Y_rd_addr  : unsigned (MEM_Y_ADDR_WIDTH-1 downto 0);
  signal mem_Y_rd_data  : signed   (MEM_Y_DATA_WIDTH-1 downto 0);

  -----------------------------------------------------------------------------
  -- signals of Cb memory
  -----------------------------------------------------------------------------
  signal mem_Cb_wr_e     : std_logic;
  signal mem_Cb_wr_addr  : unsigned (MEM_ADDR_WIDTH-1 downto 0);
  signal mem_Cb_wr_data  : signed   (MEM_DATA_WIDTH-1 downto 0);
  signal mem_Cb_rd_addr  : unsigned (MEM_ADDR_WIDTH-1 downto 0);
  signal mem_Cb_rd_data  : signed   (MEM_DATA_WIDTH-1 downto 0);

  -----------------------------------------------------------------------------
  -- signals of Cr memory
  -----------------------------------------------------------------------------
  signal mem_Cr_wr_e     : std_logic;
  signal mem_Cr_wr_addr  : unsigned (MEM_ADDR_WIDTH-1 downto 0);
  signal mem_Cr_wr_data  : signed   (MEM_DATA_WIDTH-1 downto 0);
  signal mem_Cr_rd_addr  : unsigned (MEM_ADDR_WIDTH-1 downto 0);
  signal mem_Cr_rd_data  : signed   (MEM_DATA_WIDTH-1 downto 0);


  -----------------------------------------------------------------------------
  -- Write_Signals
  -----------------------------------------------------------------------------
  type vector_size_3_t is array (0 to 2) of unsigned(1 downto 0);
  signal pixel_Y, pixel_CbCr : vector_size_3_t;

  -----------------------------------------------------------------------------
  -- Signals used to synchronize reads and writes
  -----------------------------------------------------------------------------
  type sync_state_t is (stt_sync_1, stt_sync_2, stt_sync_3, stt_sync_4, stt_sync_5, stt_sync_6);
  signal sync_state : sync_state_t;

  signal buffer_read : std_logic;
  signal buffer_wrote : std_logic;
  signal frame_wrote : std_logic;

  signal can_read : std_logic;
  signal can_write : std_logic;
  signal can_display : std_logic;

  signal s_in_request : std_logic;

begin  
  -- fixed signals
  video_clk <= clock_25Mhz;

  -- transmite immediate
  blank_n <= video_on_h and video_on_v;

mem_Y_plane: tsyncram
   generic map (
     DEPTH      => MEM_Y_DEPTH,
     DATA_WIDTH => MEM_Y_DATA_WIDTH,
     ADDR_WIDTH => MEM_Y_ADDR_WIDTH)
   port map (
     clk     => clock_25Mhz,
     wr_e    => mem_Y_wr_e,
     wr_addr => mem_Y_wr_addr,
     wr_data => mem_Y_wr_data,
     rd_addr => mem_Y_rd_addr,
     rd_data => mem_Y_rd_data);

 mem_Cb_plane: tsyncram
   generic map (
     DEPTH      => MEM_DEPTH,
     DATA_WIDTH => MEM_DATA_WIDTH,
     ADDR_WIDTH => MEM_ADDR_WIDTH)
   port map (
     clk     => clock_25Mhz,
     wr_e    => mem_Cb_wr_e,
     wr_addr => mem_Cb_wr_addr,
     wr_data => mem_Cb_wr_data,
     rd_addr => mem_Cb_rd_addr,
     rd_data => mem_Cb_rd_data);

 mem_Cr_plane: tsyncram
   generic map (
     DEPTH      => MEM_DEPTH,
     DATA_WIDTH => MEM_DATA_WIDTH,
     ADDR_WIDTH => MEM_ADDR_WIDTH)
   port map (
     clk     => clock_25Mhz,
     wr_e    => mem_Cr_wr_e,
     wr_addr => mem_Cr_wr_addr,
     wr_data => mem_Cr_wr_data,
     rd_addr => mem_Cr_rd_addr,
     rd_data => mem_Cr_rd_data);
     
----------------------------------
-- Video DAC Mode state machine --
----------------------------------
-- Set DAC in RGB mode and bias on all three channels
  process(clock_25Mhz, reset_n)
  begin
    if reset_n = '0' then
      dac_mode_current_state <= ds0;
    elsif (clock_25Mhz'event and clock_25Mhz = '1') then
      dac_mode_current_state <= dac_mode_next_state after 5 ns;
    end if;
  end process;

  process(dac_mode_current_state)
  begin

    sync_n <= '0';
    m2     <= '0';
    m1     <= '0';
    sync_t <= '0';
    case dac_mode_current_state is 

      when ds0 =>
        dac_mode_next_state <= ds1;
        
      when ds1 =>
        sync_n <= '1';
        m2     <= '0';
        m1     <= '0';
        sync_t <= '0';
        dac_mode_next_state <= ds2;

        -- bias on all three channels
      when ds2 =>
        sync_n <= '0';
        m2     <= '1';
        m1     <= '0';
        sync_t <= '0';
        dac_mode_next_state <= ds3;

      when ds3 =>
        sync_n <= '0';
        m2     <= '1';
        m1     <= '0';
        sync_t <= '0';
        dac_mode_next_state <= ds4;

        -- RGB mode
      when ds4 =>
        sync_n <= '1';
        m2     <= '0';
        m1     <= '0';
        sync_t <= '0';
        dac_mode_next_state <= ds4;
  
      when others =>
        sync_n <= '1';
        m2     <= '0';
        m1     <= '0';
        sync_t <= '0';
        dac_mode_next_state <= ds0;

    end case;
  end process;

  timing: process (clock_25Mhz, reset_n)
  begin  -- process timing
    if reset_n = '0' then               -- asynchronous reset (active low)
      count_v <= 0;
      count_h <= 0;
      vga_hs <= '1';
      vga_vs <= '1';
      video_on_h <= '1';
      video_on_v <= '1';
      column_pixel <= (others => '1');
      line_pixel <= (others => '1');
    elsif clock_25Mhz'event and clock_25Mhz = '1' then  -- rising clock edge

      if (can_display = '0') then
        count_v <= 0;
        count_h <= 0;
        vga_hs <= '1';
        vga_vs <= '1';
        video_on_h <= '1';
        video_on_v <= '1';
      else
        if (count_h = (HORIZ_RES +
                       HSYNC_BACK_PORCH_W +
                       HSYNC_FRONT_PORCH_W +
                       HSYNC_PULSE_W) - 1
            ) then
          count_h <= 0;

          if (count_v = (VERT_RES +
                         VSYNC_BACK_PORCH_W  +
                         VSYNC_FRONT_PORCH_W +
                         VSYNC_PULSE_W) - 1
              ) then
            count_v <= 0;
          else
            count_v <= count_v + 1;
          end if;
          
        else
          count_h <= count_h + 1;
        end if;

        if ((count_h >= HORIZ_RES + HSYNC_FRONT_PORCH_W) and
            (count_h < HORIZ_RES + HSYNC_FRONT_PORCH_W + HSYNC_PULSE_W)) then
          vga_hs <= '0';
        else
          vga_hs <= '1';
        end if;

        if ((count_v >= VERT_RES + VSYNC_FRONT_PORCH_W) and
            (count_v < VERT_RES + VSYNC_FRONT_PORCH_W + VSYNC_PULSE_W)) then
          vga_vs <= '0';
        else
          vga_vs <= '1';
        end if;

        -- blank_n signal 
        if count_h < HORIZ_RES then
          video_on_h <= '1';
          column_pixel <= std_logic_vector(to_unsigned(count_h,10));
        else
          video_on_h <= '0';
        end if;

        if count_v <  VERT_RES then
          video_on_v <= '1';
          line_pixel <= std_logic_vector(to_unsigned(count_v, 10));
        else
          video_on_v <= '0'; 
        end if;
      end if;
    end if;
  end process timing;


  -- purpose: Read width an height once and the frame in the YUV format
  -- type   : sequential
  -- inputs : clock_25Mhz,, reset_n
  Read: process (clock_25Mhz, reset_n)

  begin  -- process Read
    if reset_n = '0' then               -- asynchronous reset (active low)
      counter_cols <= 0;
      counter_lines <= 0;
      mem_Y_wr_e    <= '0';
      mem_Y_wr_addr <= (others => '0');
      mem_Y_wr_data <= (others => '0');
      buffer_read <= '0';
      s_in_request <= '0';
      
    elsif clock_25Mhz'event and clock_25Mhz = '1' then  -- rising clock edge

      buffer_read <= '0';
      s_in_request <= '1';
      if (can_read = '1' and in_valid = '1') then
        s_in_request <= '0';
        mem_Y_wr_e   <= '0';
        mem_Cb_wr_e   <= '0';
        mem_Cr_wr_e   <= '0';

        case read_state is
          when stt_read_height =>
            video_height <= to_integer(in_data);
            read_state <= stt_read_width;
            
          when stt_read_width =>
            video_width <= to_integer(in_data);
            read_state <= stt_read_Y;

          when stt_read_Y =>
            mem_Y_wr_e    <= '1';
            mem_Y_wr_data <= in_data;
            mem_Y_wr_addr <= resize(SHIFT_RIGHT(to_unsigned(video_width*counter_lines + counter_cols, 20), 2), MEM_Y_ADDR_WIDTH);
            
            counter_cols <= counter_cols + 4;
            if (counter_cols = video_width - 4) then  -- if it is the last four pixels
              counter_cols <= 0;
              counter_lines <= counter_lines + 1;
              if (counter_lines = video_height - 1) then
                counter_lines <= 0;
                read_state <= stt_read_Cb;
              end if;
            end if;

            
          when stt_read_Cb =>
            mem_Cb_wr_e    <= '1';
            mem_Cb_wr_data <= in_data;
            mem_Cb_wr_addr <= resize(SHIFT_RIGHT(to_unsigned((video_width/2)*counter_lines + counter_cols, 20), 2), MEM_ADDR_WIDTH);

            
            counter_cols <= counter_cols + 4;
            if (counter_cols =
                to_integer(SHIFT_RIGHT(to_unsigned(video_width, 10), 1)) - 4) then  -- if it is the last four pixels
              counter_cols <= 0;
              counter_lines <= counter_lines + 1;
              assert (not ((counter_lines mod 40) = 0)) report "Writing line:"&integer'image(counter_lines) severity note;
              if (counter_lines =
                  (to_integer(SHIFT_RIGHT(to_unsigned(video_height, 10), 1))- 1)) then
                counter_lines <= 0;
                read_state <= stt_read_Cr;
              end if;
            end if;

            
          when stt_read_Cr =>
            mem_Cr_wr_e    <= '1';
            mem_Cr_wr_data <= in_data;
            mem_Cr_wr_addr <= resize(SHIFT_RIGHT(to_unsigned((video_width/2)*counter_lines + counter_cols, 20), 2), MEM_ADDR_WIDTH);

            
            counter_cols <= counter_cols + 4;
            if (counter_cols =
                to_integer(SHIFT_RIGHT(to_unsigned(video_width, 10), 1)) - 4) then  -- if it is the last four pixels
              counter_cols <= 0;
              counter_lines <= counter_lines + 1;
              if (counter_lines =
                  (to_integer(SHIFT_RIGHT(to_unsigned(video_height, 10), 1))- 1)) then
                counter_lines <= 0;
                read_state <= stt_read_Y;
                buffer_read <= '1';
              end if;
            end if;
        end case;
      end if;


      
    end if;
  end process Read;
  

  display_image: process(clock_25Mhz, reset_n)
    variable var_Y, var_Cb, var_Cr : unsigned(7 downto 0);
    variable v_red, v_green, v_blue : signed(31 downto 0);
    variable v_addr_Y, v_addr_CbCr : unsigned(19 downto 0);
    variable temp : signed(31 downto 0);

  begin  -- process display_image

    if reset_n = '0' then               -- asynchronous reset (active low)
      mem_Y_rd_addr  <= (others => '0');
      mem_Cb_rd_addr <= (others => '0');
      mem_Cr_rd_addr <= (others => '0');

      mem_read_col <= 0;
      mem_read_line <= 0;
      buffer_wrote <= '0';
      frame_wrote <= '0';
      red <= x"ff";
      green <= x"ff";
      blue <= x"ff";

      pixel_Y(0) <= "00";
      pixel_Y(1) <= "00";

      pixel_CbCr(0) <= "00";
      pixel_CbCr(1) <= "00";

      write_state <= stt_write_data;
      
    elsif clock_25Mhz'event and clock_25Mhz = '1' then  -- rising clock edge

      buffer_wrote <= '0';
      frame_wrote <= '0';
      
      red <= x"ff";
      green <= x"ff";
      blue <= x"ff";

--       assert (not (mem_read_col = 50 and mem_read_line = 1)) report "Fim" severity FAILURE;
      
      if (can_write = '1') then
        v_addr_Y := to_unsigned(mem_read_line * video_width + mem_read_col, 20);
        v_addr_CbCr := resize(SHIFT_RIGHT(to_unsigned(mem_read_line, 10), 1) *
                              SHIFT_RIGHT(to_unsigned(video_width, 10), 1) +
                              SHIFT_RIGHT(to_unsigned(mem_read_col, 10), 1), 20);
        
        mem_Y_rd_addr  <= resize(SHIFT_RIGHT(v_addr_Y, 2), MEM_Y_ADDR_WIDTH);
        mem_Cb_rd_addr <= resize(SHIFT_RIGHT(v_addr_CbCr, 2), MEM_ADDR_WIDTH);
        mem_Cr_rd_addr <= resize(SHIFT_RIGHT(v_addr_CbCr, 2), MEM_ADDR_WIDTH);

        pixel_Y(0) <= pixel_Y(1);
        pixel_Y(1) <= v_addr_Y(1 downto 0);

        pixel_CbCr(0) <= pixel_CbCr(1);
        pixel_CbCr(1) <= v_addr_CbCr(1 downto 0);

        assert false report "v_addr_Y="&integer'image(to_integer(v_addr_Y)) severity note;
        assert false report "v_addr_CbCr="&integer'image(to_integer(v_addr_CbCr)) severity note;
        
        assert false report "pixel_Y(0)="&integer'image(to_integer(pixel_Y(0))) severity note;
        assert false report "pixel_Y(1)="&integer'image(to_integer(pixel_Y(1))) severity note;
        assert false report "pixel_Y(2)="&integer'image(to_integer(pixel_Y(1))) severity note;


        assert false report "pixel_CbCr(0)="&integer'image(to_integer(pixel_CbCr(0))) severity note;
        assert false report "pixel_CbCr(1)="&integer'image(to_integer(pixel_CbCr(1))) severity note;
        assert false report "pixel_CbCr(2)="&integer'image(to_integer(pixel_CbCr(2))) severity note;

        assert false report "v_addr_Y="&integer'image(to_integer(v_addr_Y)) severity note;
        assert false report "v_addr_CbCr="&integer'image(to_integer(v_addr_CbCr)) severity note;

        assert false report "mem_read_col="&integer'image(mem_read_col) severity note;
        assert false report "mem_read_line="&integer'image(mem_read_line) severity note;

        
        case pixel_Y(0) is
          when "00" =>
            var_Y := unsigned(mem_Y_rd_data(31 downto 24));
          when "01" =>
            var_Y := unsigned(mem_Y_rd_data(23 downto 16));
          when "10" =>
            var_Y := unsigned(mem_Y_rd_data(15 downto 8));
          when "11" =>
            var_Y := unsigned(mem_Y_rd_data(7 downto 0));
          when others => null;
        end case;

        

        case pixel_CbCr(0) is
          when "00" =>
            var_Cb := unsigned(mem_Cb_rd_data(31 downto 24));
            var_Cr := unsigned(mem_Cr_rd_data(31 downto 24));
          when "01" =>
            var_Cb := unsigned(mem_Cb_rd_data(23 downto 16));
            var_Cr := unsigned(mem_Cr_rd_data(23 downto 16));
          when "10" =>
            var_Cb := unsigned(mem_Cb_rd_data(15 downto 8));
            var_Cr := unsigned(mem_Cr_rd_data(15 downto 8));
          when "11" =>
            var_Cb := unsigned(mem_Cb_rd_data(7 downto 0));
            var_Cr := unsigned(mem_Cr_rd_data(7 downto 0));
          when others => null;
        end case;

--          assert false report "var_Y="&integer'image(to_integer(var_Y)) severity note;
--          assert false report "var_Cb="&integer'image(to_integer(var_Cb)) severity note;
--          assert false report "var_Cr="&integer'image(to_integer(var_Cr)) severity note;
--          assert false report "--end pixel" severity note;
--          assert false report "" severity note;
        
        v_red := resize(SHIFT_RIGHT (to_signed(1220542, 22) * (signed('0' & var_Y) - 16) +
                                     to_signed(1673527, 22) * (signed('0' & var_Cr) - 128), 20), 32);

        v_green := resize(SHIFT_RIGHT(to_signed(1220542, 22) * (signed('0' &var_Y) - 16) -
                                      to_signed(852492, 22) * (signed('0' &var_Cr) - 128) -
                                      to_signed(411042, 22) *(signed('0' &var_Cb) - 128), 20), 32);

        v_blue := resize(SHIFT_RIGHT (to_signed(1220542, 22) * (signed('0' &var_Y) - 16) +
                                      to_signed(2114978, 23) * (signed('0' &var_Cb) - 128), 20), 32);

        if (v_red < 0) then
          red <= x"00";
        elsif (v_red > 255) then
          red <= x"ff";
        else
          red <= std_logic_vector(v_red(7 downto 0));
        end if;

        if (v_green < 0) then
          green <= x"00";
        elsif (v_green > 255) then
          green <= x"ff";
        else
          green <= std_logic_vector(v_green(7 downto 0));
        end if;

        if (v_blue < 0) then
          blue <= x"00";
        elsif (v_blue > 255) then
          blue <= x"ff";
        else
          blue <= std_logic_vector(v_blue(7 downto 0));
        end if;

        case write_state is
          when stt_write_data => 
            mem_read_col <= mem_read_col + 1;
            if (mem_read_col = video_width - 1) then  -- if it is the last pixel
              write_state <= stt_write_wait1;
              mem_read_col <= 0;
              mem_read_line <= mem_read_line + 1;
              if (mem_read_line = video_height - 1) then
                mem_read_line <= 0;
                buffer_wrote <= '1';
              end if;
            end if;

          when stt_write_wait1 =>
            write_state <= stt_write_wait2;
            
          when stt_write_wait2 =>
            write_state <= stt_write_wait3;

          when others =>
            write_state <= stt_write_none;
            red <= x"00";
            green <= x"00";
            blue <= x"00";

            if (count_h = (HORIZ_RES + HSYNC_BACK_PORCH_W + HSYNC_FRONT_PORCH_W + HSYNC_PULSE_W) - 2) then
              if (count_v < video_height - 1) then
                mem_read_col <= mem_read_col + 1;
                write_state <= stt_write_data;
              elsif (count_v = (VERT_RES + VSYNC_BACK_PORCH_W + VSYNC_FRONT_PORCH_W + VSYNC_PULSE_W) - 1) then
                mem_read_col <= mem_read_col + 1;
                write_state <= stt_write_data;
                frame_wrote <= '1';
                assert false report "Frame wrote" severity note;
              end if;
            end if;

        end case;
      end if;  
    end if;
  end process display_image;

  in_request <= can_read;
  
  -- purpose: Synchronize the Timing, Read and Display_Image process
  -- type   : combinational
  -- inputs : sync_state , record_state, reset_n, write_state, s_in_request
  -- outputs: can_write, can_read, can_display
  synchronizer: process (reset_n, sync_state, s_in_request,
                        write_state, clock_25Mhz,
                        video_on_h, video_on_v)
  begin  -- process synchronizer
    if reset_n = '0' then
      can_write <= '0';
      can_display <= '0';
      can_read <= '0';

      --remover depois
      out_valid <= '0';
    else
    
      out_valid <= '0';
      case sync_state is
        when stt_sync_1 =>
          can_display <= '0';
          can_read <= s_in_request;
          can_write <= '0';
          
        when stt_sync_2 =>
          can_display <= '0';
          can_read <= '0';
          can_write <= '1';

        when stt_sync_3 =>
          can_display <= '1';
          can_read <= '0';
          can_write <= '1';

          if (write_state /= stt_write_none) then
            out_valid <= clock_25Mhz and video_on_h and video_on_v;
          else
            out_valid <= '0';
          end if;

        when stt_sync_4 =>
          can_display <= '1';
          can_read <= s_in_request;
          can_write <= '1';

        when stt_sync_5 =>
          can_display <= '1';
          can_read <= s_in_request;
          can_write <= '1';

        when stt_sync_6 =>
          can_display <= '1';
          can_read <= '0';
          can_write <= '1';

      end case;
    end if;
  end process synchronizer;

  -- purpose: Synchronize the Timing, Read and Display_Image process
  -- type   : sequential
  -- inputs : clock_25Mhz,, reset_n, buffer_read, buffer_wrote, frame_wrote,
  --          mem_read_line, mem_read_col, sync_state 
  -- outputs: sync_state
  synchronizerControl: process (clock_25Mhz, reset_n)
  begin  -- process synchronizerControl
    if reset_n = '0' then               -- asynchronous reset (active low)
      sync_state <= stt_sync_1;
    elsif clock_25Mhz'event and clock_25Mhz = '1' then  -- rising clock edge
      sync_state <= sync_state;
      case sync_state is
        -----------------------------------------------------------------------
        -- Initial state, wait until the buffer been completely written
        -----------------------------------------------------------------------
        when stt_sync_1 =>
          -- Transition
          if (buffer_read = '1') then
            sync_state <= stt_sync_2;
          end if;

        -----------------------------------------------------------------------
        -- Synchronize the delay of memory with the output
        -----------------------------------------------------------------------
        when stt_sync_2 =>
          -- Transition
          if (mem_read_col = 1 and mem_read_line = 0) then
            sync_state <= stt_sync_3;
          end if;

        -----------------------------------------------------------------------
        -- Display the buffer the first time
        -----------------------------------------------------------------------
        when stt_sync_3 =>
          -- Transition
          if (buffer_wrote = '1') then
            sync_state <= stt_sync_4;
          end if;

        -----------------------------------------------------------------------
        -- Complete the pixel sweep
        -----------------------------------------------------------------------
        when stt_sync_4 =>
          -- Transition
          if (frame_wrote = '1' and buffer_read = '1') then
            sync_state <= stt_sync_3;

          elsif (frame_wrote = '1') then
            sync_state <= stt_sync_5;

          elsif (buffer_read = '1') then
            sync_state <= stt_sync_6;
          end if;

        -----------------------------------------------------------------------
        -- Wait until the buffer been read
        -----------------------------------------------------------------------
        when stt_sync_5 =>
          -- Transition
          if (buffer_read = '1') then
            sync_state <= stt_sync_3;
          end if;

        -----------------------------------------------------------------------
        -- Wait until the frame been complety displayed
        -----------------------------------------------------------------------
        when stt_sync_6 =>
          -- Transition
          if (frame_wrote = '1') then
            sync_state <= stt_sync_3;
          end if;

      end case;
    end if;
  end process synchronizerControl;


end a_lancelot;
