library std;
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library gaisler;
use gaisler.theora_apb.all;

entity theora_hardware is
  
  port (
    clk : in std_logic;
    clk_25Mhz : in std_logic;
    reset_n      : in  std_logic;                     -- reset

    ---------------------------------------------------------------------------
    -- Ports of Handshake
    ---------------------------------------------------------------------------
    in_request    : out std_logic;
    in_valid      : in  std_logic;      -- in_data
    in_data       : in  signed(31 downto 0);

    -- Remover depois
    out_valid     : out std_logic;
    ---------------------------------------------------------------------------
    -- Ports of video controller
    ---------------------------------------------------------------------------
    red	         : out std_logic_vector(7 downto 0);  -- red component
    green	 : out std_logic_vector(7 downto 0);  -- green component
    blue	 : out std_logic_vector(7 downto 0);  -- blue component
    line_pixel   : out std_logic_vector(9 downto 0);  -- compute line
    column_pixel : out std_logic_vector(9 downto 0);  -- compute column
    m1, m2       : out std_logic;                     -- select dac mode
    blank_n      : out std_logic;                     -- dac command
    sync_n       : out std_logic;                     -- dac command
    sync_t       : out std_logic;                     -- dac command
    video_clk    : out std_logic;                     -- dac command
    vga_vs       : out std_logic;                     -- vertical sync
    vga_hs       : out std_logic                      -- horizontal sync
    );

end theora_hardware;

architecture a_theora_hardware of theora_hardware is


  component interface_vga
    port(
      clock_25Mhz  : in  std_logic;
      reset_n      : in  std_logic;                     -- reset

      ---------------------------------------------------------------------------
      -- Ports of Handshake
      ---------------------------------------------------------------------------
      in_request    : out std_logic;
      in_valid      : in  std_logic;      -- in_data
      in_data       : in  signed(31 downto 0);

      -- Remover depois
      out_valid     : out std_logic;
      ---------------------------------------------------------------------------
      -- Ports of video controller
      ---------------------------------------------------------------------------

      red	         : out std_logic_vector(7 downto 0);  -- red component
      green	 : out std_logic_vector(7 downto 0);  -- green component
      blue	 : out std_logic_vector(7 downto 0);  -- blue component
      line_pixel   : out std_logic_vector(9 downto 0);  -- compute line
      column_pixel : out std_logic_vector(9 downto 0);  -- compute column
      m1, m2       : out std_logic;                     -- select dac mode
      blank_n      : out std_logic;                     -- dac command
      sync_n       : out std_logic;                     -- dac command
      sync_t       : out std_logic;                     -- dac command
      video_clk    : out std_logic;                     -- dac command
      vga_vs       : out std_logic;                     -- vertical sync
      vga_hs       : out std_logic                      -- horizontal sync
      );

  end component;

  component ReconRefFrames
    port (Clk,
          Reset_n       : in  std_logic;
          
          in_request    : out std_logic;
          in_valid      : in  std_logic;
          in_data       : in  signed(31 downto 0);
          
          out_requested : in  std_logic;
          out_valid     : out std_logic;
          out_data      : out signed(31 downto 0)
          );

  end component;

  signal out_iv_requested : std_logic;
  signal out_iv_valid     : std_logic;
  signal out_iv_data      : signed(31 downto 0);

  signal in_rr_request    : std_logic;
  signal in_rr_valid      : std_logic;
  signal in_rr_data       : signed(31 downto 0);
  
  signal out_rr_requested : std_logic;
  signal out_rr_valid     : std_logic;
  signal out_rr_data      : signed(31 downto 0);

  type switch_state_t is (stt_switch1, stt_switch2, stt_switch3,
                          stt_switch4, stt_switch5);
  signal switch_state : switch_state_t;
  signal s_in_request : std_logic;

  
  signal s_out_iv_requested : std_logic;
  signal s_out_iv_valid     : std_logic;
  signal s_out_iv_data      : signed(31 downto 0);

  type handshake_state_t is (stt_handshake1, stt_handshake2, stt_handshake3, stt_handshake4);
  signal handshake_state : handshake_state_t;

  signal video_height, video_width : integer range 0 to 1023;
  signal data_source : unsigned(1 downto 0);
  
  --signal clk_25Mhzz : std_logic;
begin  -- a_theora_hardware

  interface_vga0: interface_vga
    port map (
      clock_25Mhz  => clk_25Mhz,
      reset_n      => reset_n,

      ---------------------------------------------------------------------------
      -- Ports of Handshake
      ---------------------------------------------------------------------------
      in_request    => out_iv_requested,
      in_valid      => out_iv_valid,
      in_data       => out_iv_data,

      -- Remover depois
      out_valid     => out_valid,
      ---------------------------------------------------------------------------
      -- Ports of video controller
      ---------------------------------------------------------------------------
      red	   => red,
      green	   => green,
      blue	   => blue,
      line_pixel   => line_pixel,
      column_pixel => column_pixel,
      m1           => m1,
      m2           => m2,
      blank_n      => blank_n,
      sync_n       => sync_n,
      sync_t       => sync_t,
      video_clk    => video_clk,
      vga_vs       => vga_vs,
      vga_hs       => vga_hs
      );

  reconrefframes0: ReconRefFrames
    port map (
      Clk => clk,
      Reset_n       => reset_n,
        
      in_request    => out_rr_requested,
      in_valid      => out_rr_valid,
      in_data       => out_rr_data,
        
      out_requested => in_rr_request,
      out_valid     => in_rr_valid,
      out_data      => in_rr_data
      );

  in_request <= s_in_request;
  out_rr_valid <= in_valid;
  out_rr_data <= in_data;


  HandshakeSendControl: process (handshake_state, s_out_iv_valid, s_out_iv_data)
  begin  -- process HandshakeSendControl
    if (handshake_state = stt_handshake1 or handshake_state = stt_handshake2) then
      out_iv_valid <= s_out_iv_valid;
      out_iv_data <= s_out_iv_data;

    else
      out_iv_valid <= '0';
      out_iv_data <= (others => '0');
    end if;
  end process HandshakeSendControl;

  
  SwitchSignalsIn: process (reset_n,
                            in_valid, in_data, s_out_iv_requested,
                            in_rr_valid, in_rr_data,
                            out_iv_requested, out_rr_requested,
                            switch_state, handshake_state)
  begin  -- process SwitchSignals

    s_in_request <= '0';
    in_rr_request <= '0';
    
    if (switch_state = stt_switch3 or  switch_state = stt_switch4) then
      in_rr_request <= '0';
      s_in_request <= out_rr_requested and s_out_iv_requested;

    else
      in_rr_request <= s_out_iv_requested;    
      s_in_request <= out_rr_requested;

    end if;

  end process SwitchSignalsIn;


  s_out_iv_requested <= out_iv_requested when handshake_state = stt_handshake4 else
                        '0';
                        
  -- purpose: Control the switch the interface_vga and ReconRefFrames ports' signals
  -- type   : sequential
  -- inputs : clk, reset_n
  SwitchSignalsControl: process (clk, reset_n)
  begin  -- process SwitchEntries
    if reset_n = '0' then               -- asynchronous reset (active low)
      switch_state <= stt_switch1;
      --clk_25Mhzz <= '0';
    elsif clk'event and clk = '1' then  -- rising clock edge
    
    --clk_25Mhzz <= not clk_25Mhzz;
      if (s_in_request = '1' and in_valid = '1') then
        case switch_state is
          when stt_switch1 =>
            switch_state <= stt_switch2;
            
          when stt_switch2 =>
            switch_state <= stt_switch3;

          when stt_switch3 =>
            switch_state <= stt_switch4;
            video_height <= to_integer(in_data);
            
          when stt_switch4 =>
            switch_state <= stt_switch5;
            video_width <= to_integer(in_data);

          when stt_switch5 =>
            switch_state <= stt_switch5;
        end case;
      end if;
    end if;
  end process SwitchSignalsControl;

      
  HandshakeSendTransitions: process (clk, reset_n)
  begin  -- process SwitchEntries
    if reset_n = '0' then               -- asynchronous reset (active low)
      handshake_state <= stt_handshake4;
      s_out_iv_valid <= '0';
      s_out_iv_data <= (others => '0');
      data_source <= "00";
      
      
    elsif clk'event and clk = '1' then  -- rising clock edge
      
      if switch_state = stt_switch5 then
        if handshake_state = stt_handshake1 then
          if (out_iv_requested = '1') then
            handshake_state <= stt_handshake2;
          end if;
          
        elsif (handshake_state = stt_handshake2) then
          if (out_iv_requested = '0') then
            handshake_state <= stt_handshake3;
            s_out_iv_valid <= '0';
            s_out_iv_data <= (others => '0');
          end if;

          -- out_iv_valid <= '0'
        elsif (handshake_state = stt_handshake3) then
          if (out_iv_requested = '1') then
            handshake_state <= stt_handshake4;
          end if;

        else

          if (data_source = "00") then
            s_out_iv_valid <= '1';
            s_out_iv_data <= to_signed(video_height, 32);
            data_source <= data_source + 1;
            handshake_state <= stt_handshake1;

          elsif (data_source = "01") then
            s_out_iv_valid <= '1';
            s_out_iv_data <= to_signed(video_width, 32);
            data_source <= data_source + 1;
            handshake_state <= stt_handshake1;
          else
            if (out_iv_requested = '1' and in_rr_valid = '1') then
              s_out_iv_valid <= in_rr_valid;
              s_out_iv_data <= in_rr_data;
              handshake_state <= stt_handshake1;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process HandshakeSendTransitions;
end a_theora_hardware;
